@@ -39,6 +39,7 @@ using System.Collections.Generic;
 using System.Collections.ObjectModel;
 using System.Data;
 using System.IO;
+using System.Linq;
 using System.Net;
 using System.Runtime.Serialization;
 using System.Text;
@@ -1428,6 +1429,27 @@ namespace MonoTests.System.Runtime.Serialization
 			Assert.IsTrue (idx2 < 0, "idx2 should not occur at " + idx2);
 		}
 
+		[Test]
+		public void AncestralReference ()
+		{
+			// Reference to Parent comes inside the Parent itself.
+			// In this case, adding reference after complete deserialization won't work (but it should).
+			var ms = new MemoryStream ();
+			Type [] knownTypes = new Type [] { typeof (ParentClass), typeof (Foo), typeof (Bar) };
+
+            		var ds = new DataContractSerializer (typeof (Parent));
+
+			var org = new Parent ();
+			ds.WriteObject (ms, org);
+			string result = Encoding.UTF8.GetString (ms.ToArray ());
+			ms.Position = 0;
+
+			var parent = (Parent) ds.ReadObject (ms);
+
+			Assert.IsNotNull (parent.Child, "#1");
+			Assert.AreEqual (parent, parent.Child.Parent, "#2");
+		}
+
 		[Test]
 		public void IXmlSerializableCallConstructor ()
 		{
@@ -1478,6 +1500,81 @@ namespace MonoTests.System.Runtime.Serialization
 			
 			var ds = (DataSet) x.ReadObject (r);
 		}
+
+		[Test]
+		[ExpectedException (typeof (InvalidDataContractException))] // BaseConstraintType1 is neither DataContract nor Serializable.
+		public void BaseConstraint1 ()
+		{
+			new DataContractSerializer (typeof (BaseConstraintType3)).WriteObject (XmlWriter.Create (TextWriter.Null), new BaseConstraintType3 ());
+		}
+
+		[Test]
+		public void BaseConstraint2 ()
+		{
+			new DataContractSerializer (typeof (BaseConstraintType4)).WriteObject (XmlWriter.Create (TextWriter.Null), new BaseConstraintType4 ());
+		}
+
+		[Test] // bug #652331
+		public void MembersNamespacesInBaseType ()
+		{
+			string xml1 = @"<Currency>JPY</Currency><Description i:nil=""true"" />";
+			string xml2 = @"<Currency xmlns=""http://schemas.datacontract.org/2004/07/SLProto5"">JPY</Currency><Description i:nil=""true"" xmlns=""http://schemas.datacontract.org/2004/07/SLProto5"" />";
+			Assert.AreEqual ("JPY", MembersNamespacesInBaseType_Part<SLProto5.CashAmount> (new SLProto5.CashAmount () { Currency = "JPY" }, xml1, "#1").Currency, "r#1");
+			Assert.AreEqual ("JPY", MembersNamespacesInBaseType_Part<SLProto5_Different.CashAmount> (new SLProto5_Different.CashAmount () { Currency = "JPY" }, xml2, "#2").Currency, "r#2");
+			Assert.AreEqual ("JPY", MembersNamespacesInBaseType_Part<SLProto5.CashAmountDC> (new SLProto5.CashAmountDC () { Currency = "JPY" }, xml1, "#3").Currency, "r#3");
+			Assert.AreEqual ("JPY", MembersNamespacesInBaseType_Part<SLProto5_Different.CashAmountDC> (new SLProto5_Different.CashAmountDC () { Currency = "JPY" }, xml2, "#4").Currency, "r#4");
+		}
+
+		T MembersNamespacesInBaseType_Part<T> (T instance, string expectedPart, string assert)
+		{
+			var ds = new DataContractSerializer (typeof (T));
+			var sw = new StringWriter ();
+			using (var w = XmlWriter.Create (sw))
+				ds.WriteObject (w, instance);
+			Assert.IsTrue (sw.ToString ().IndexOf (expectedPart) > 0, assert);
+			return (T) ds.ReadObject (XmlReader.Create (new StringReader (sw.ToString ())));
+		}
+
+		[Test]
+		public void DateTimeOffsetSerialization ()
+		{
+			var ds = new DataContractSerializer (typeof (DateTimeOffset));
+			var sw = new StringWriter ();
+			string xml = "<DateTimeOffset xmlns:i='http://www.w3.org/2001/XMLSchema-instance' xmlns='http://schemas.datacontract.org/2004/07/System'><DateTime>2011-03-01T02:05:06.078Z</DateTime><OffsetMinutes>120</OffsetMinutes></DateTimeOffset>".Replace ('\'', '"');
+			var v = new DateTimeOffset (new DateTime (2011, 3, 1, 4, 5, 6, 78), TimeSpan.FromMinutes (120));
+			using (var xw = XmlWriter.Create (sw, settings)) {
+				ds.WriteObject (xw, v);
+			}
+			Assert.AreEqual (xml, sw.ToString (), "#1");
+			Assert.AreEqual (v, ds.ReadObject (XmlReader.Create (new StringReader (sw.ToString ()))), "#2");
+		}
+
+		[Test]
+		public void XmlDocumentSupport ()
+		{
+			var xml = "<XmlDocumentContract xmlns:i='http://www.w3.org/2001/XMLSchema-instance' xmlns='urn:foo'><Content><Root xmlns=''>Hello, world!</Root></Content><Nodes><child1 xmlns='' /><child2 xmlns='' /></Nodes></XmlDocumentContract>".Replace ('\'', '"');
+			var xml2 = "<Root>Hello, world!</Root>";
+			var obj = new XmlDocumentContract ();
+			var doc = new XmlDocument ();
+			doc.LoadXml (xml2);
+			obj.Content = doc.DocumentElement;
+			doc = new XmlDocument ();
+			doc.LoadXml ("<root><child1/><child2/></root>");
+			var l = new List<XmlNode> ();
+			foreach (XmlNode node in doc.DocumentElement.ChildNodes)
+				l.Add (node);
+			obj.Nodes = l.ToArray ();
+			var serializer = new DataContractSerializer (typeof (XmlDocumentContract))
+			;
+			var sb = new StringBuilder ();
+			using (var writer = new StringWriter (sb))
+				serializer.WriteObject (new XmlTextWriter (writer), obj);
+			Assert.AreEqual (xml, sb.ToString (), "#1");
+			using (var reader = new StringReader (sb.ToString ()))
+				obj = serializer.ReadObject (new XmlTextReader (reader)) as XmlDocumentContract;
+			Assert.AreEqual ("Hello, world!", obj.Content != null ? obj.Content.InnerText : String.Empty, "#2");
+			Assert.AreEqual (2, obj.Nodes != null ? obj.Nodes.Length : -1, "#3");
+		}
 	}
 	
 	[DataContract]
@@ -1779,6 +1876,45 @@ namespace MonoTests.System.Runtime.Serialization
 		[DataMember]
 		public string X = "x";
 	}
+
+	class BaseConstraintType1 // non-serializable
+	{
+	}
+	
+	[Serializable]
+	class BaseConstraintType2
+	{
+	}
+	
+	[DataContract]
+	class BaseConstraintType3 : BaseConstraintType1
+	{
+	}
+	
+	[DataContract]
+	class BaseConstraintType4 : BaseConstraintType2
+	{
+	}
+
+	[DataContract (Namespace = "urn:foo")]
+	public class XmlDocumentContract
+	{
+		[DataMember (Name = "Content")]
+		private XmlElement content;
+
+		public XmlElement Content {
+			get { return content; }
+			set { content = value; }
+		}
+
+		[DataMember (Name = "Nodes")]
+		private XmlNode [] nodes;
+
+		public XmlNode [] Nodes {
+			get { return nodes; }
+			set { nodes = value; }
+		}
+	}
 }
 
 [DataContract]
@@ -1954,4 +2090,73 @@ public class Dict<T> : Dictionary<string, T> where T : ParentClass
 	}
 	
 }
+
+[DataContract (IsReference = true)]
+public class Parent
+{
+	//constructor
+	public Parent ()
+	{
+		Child = new Child (this);
+	}
+
+	[DataMember]
+	public Child Child;
+}
+
+[DataContract]
+public class Child
+{
+	public Child ()
+	{
+	}
+	
+	public Child (Parent parent)
+	{
+		this.Parent = parent;
+	}
+	
+	[DataMember]
+	public Parent Parent;
+}
+
+namespace SLProto5
+{
+	public class CashAmount : Amount
+	{
+	}
+
+	[DataContract]
+	public class CashAmountDC : AmountDC
+	{
+	}
+
+	public class Amount
+	{
+		public string Currency { get; set; }
+		public string Description { get; set; }
+	}
+
+	[DataContract]
+	public class AmountDC
+	{
+		[DataMember]
+		public string Currency { get; set; }
+		[DataMember]
+		public string Description { get; set; }
+	}
+}
+
+namespace SLProto5_Different
+{
+	public class CashAmount : SLProto5.Amount
+	{
+	}
+
+	[DataContract]
+	public class CashAmountDC : SLProto5.AmountDC
+	{
+	}
+}
+
 #endregion
