@@ -47,6 +47,7 @@ using System.Web.Script.Serialization;
 using System.Web.Script.Services;
 using System.Xml;
 using System.Collections.ObjectModel;
+using System.Web.Util;
 
 namespace System.Web.UI
 {
@@ -64,7 +65,9 @@ namespace System.Web.UI
 		const string hiddenField = "hiddenField";
 		const string arrayDeclaration = "arrayDeclaration";
 		const string scriptBlock = "scriptBlock";
+#if NET_3_5
 		const string scriptStartupBlock = "scriptStartupBlock";
+#endif
 		const string expando = "expando";
 		const string onSubmit = "onSubmit";
 		const string asyncPostBackControlIDs = "asyncPostBackControlIDs";
@@ -235,8 +238,7 @@ namespace System.Web.UI
 				if (IsDeploymentRetail)
 					return false;
 
-				CompilationSection compilation = (CompilationSection) WebConfigurationManager.GetSection ("system.web/compilation");
-				if (!compilation.Debug && (ScriptMode == ScriptMode.Auto || ScriptMode == ScriptMode.Inherit))
+				if (!RuntimeHelpers.DebuggingEnabled && (ScriptMode == ScriptMode.Auto || ScriptMode == ScriptMode.Inherit))
 					return false;
 
 				if (ScriptMode == ScriptMode.Release)
@@ -400,7 +402,7 @@ namespace System.Web.UI
 			return GetCurrentInternal (page);
 		}
 
-		static ScriptManager GetCurrentInternal (Page page)
+		internal static ScriptManager GetCurrentInternal (Page page)
 		{
 			if (page == null)
 				return null;
@@ -453,7 +455,7 @@ namespace System.Web.UI
 				AsyncPostBackError (this, e);
 		}
 
-		protected override void OnInit (EventArgs e) {
+		protected internal override void OnInit (EventArgs e) {
 			base.OnInit (e);
 
 			if (GetCurrentInternal (Page) != null)
@@ -469,7 +471,7 @@ namespace System.Web.UI
 				OnAsyncPostBackError (new AsyncPostBackErrorEventArgs (Context.Error));
 		}
 
-		protected override void OnPreRender (EventArgs e) {
+		protected internal override void OnPreRender (EventArgs e) {
 			base.OnPreRender (e);
 
 			Page.PreRenderComplete += new EventHandler (OnPreRenderComplete);
@@ -1084,7 +1086,7 @@ namespace System.Web.UI
 			scriptList.Add (new RegisteredScript (control, type, key, script, url, addScriptTag, scriptType));
 		}
 
-		protected override void Render (HtmlTextWriter writer) {
+		protected internal override void Render (HtmlTextWriter writer) {
 			// MSDN: This method is used by control developers to extend the ScriptManager control. 
 			// Notes to Inheritors: 
 			// When overriding this method, call the base Render(HtmlTextWriter) method 
@@ -1190,17 +1192,16 @@ namespace System.Web.UI
 
 		#endregion
 
-		internal void WriteCallbackException (TextWriter output, Exception ex, bool writeMessage) {
+		internal static void WriteCallbackException (ScriptManager current, TextWriter output, Exception ex, bool writeMessage)
+		{
 #if TARGET_DOTNET
 			if (ex is HttpUnhandledException)
 				ex = ex.InnerException;
 #endif
 			HttpException httpEx = ex as HttpException;
-			string message = AsyncPostBackErrorMessage;
+			string message = current != null ? current.AsyncPostBackErrorMessage : null;
 			if (String.IsNullOrEmpty (message) && writeMessage) {
-				HttpContext ctx = HttpContext.Current;
-				
-				if (IsDebuggingEnabled)
+				if ((current != null && current.IsDebuggingEnabled) || (current == null && RuntimeHelpers.DebuggingEnabled))
 					message = ex.ToString ();
 				else
 					message = ex.Message;
@@ -1404,12 +1405,14 @@ namespace System.Web.UI
 							else
 								WriteCallbackOutput (output, scriptBlock, scriptContentWithTags, SerializeScriptBlock (scriptEntry));
 							break;
+#if NET_3_5
 						case RegisteredScriptType.ClientStartupScript:
 							if (scriptEntry.AddScriptTags)
 								WriteCallbackOutput (output, scriptStartupBlock, scriptContentNoTags, scriptEntry.Script);
 							else
 								WriteCallbackOutput (output, scriptStartupBlock, scriptContentWithTags, SerializeScriptBlock (scriptEntry));
 							break;
+#endif
 						case RegisteredScriptType.ClientScriptInclude:
 							WriteCallbackOutput (output, scriptBlock, scriptPath, scriptEntry.Url);
 							break;
