@@ -3,9 +3,10 @@
 //
 // Authors:
 //	Gonzalo Paniagua Javier (gonzalo@ximian.com)
+//      Marek Habersack <mhabersack@novell.com>
 //
 // (C) 2002,2003 Ximian, Inc (http://www.ximian.com)
-// Copyright (c) 2004,2006 Novell, Inc (http://www.novell.com)
+// Copyright (c) 2004-2009 Novell, Inc (http://www.novell.com)
 //
 
 //
@@ -30,6 +31,7 @@
 //
 using System;
 using System.Collections;
+using System.Collections.Generic;
 using System.CodeDom.Compiler;
 using System.Globalization;
 using System.IO;
@@ -52,7 +54,7 @@ namespace System.Web.Compilation
 		public BuilderLocation (ControlBuilder builder, ILocation location)
 		{
 			this.Builder = builder;
-			this.Location = location;
+			this.Location = new Location (location);
 		}
 	}
 
@@ -165,7 +167,7 @@ namespace System.Web.Compilation
 			if (tags.Count == 0)
 				return false;
 
-			return 0 == String.Compare (tagid, (string) tags.Peek (), true, CultureInfo.InvariantCulture);
+			return 0 == String.Compare (tagid, (string) tags.Peek (), true, Helpers.InvariantCulture);
 		}
 		
 		public int Count {
@@ -177,8 +179,58 @@ namespace System.Web.Compilation
 		}
 	}
 
+	enum TextBlockType
+	{
+		Verbatim,
+		Expression,
+		Tag,
+		Comment
+	}
+	
+	sealed class TextBlock
+	{
+		public string Content;
+		public readonly TextBlockType Type;
+		public readonly int Length;
+		
+		public TextBlock (TextBlockType type, string content)
+		{
+			Content = content;
+			Type = type;
+			Length = content.Length;
+		}
+
+		public override string ToString ()
+		{
+			return this.GetType ().FullName + " [" + this.Type + "]";
+		}
+	}
+	
 	class AspGenerator
 	{
+		const int READ_BUFFER_SIZE = 8192;
+		
+		internal static Regex DirectiveRegex = new Regex (@"<%\s*@(\s*(?<attrname>\w[\w:]*(?=\W))(\s*(?<equal>=)\s*""(?<attrval>[^""]*)""|\s*(?<equal>=)\s*'(?<attrval>[^']*)'|\s*(?<equal>=)\s*(?<attrval>[^\s%>]*)|(?<equal>)(?<attrval>\s*?)))*\s*?%>", RegexOptions.Compiled | RegexOptions.IgnoreCase);
+
+		static readonly Regex runatServer = new Regex (@"<[\w:\.]+.*?runat=[""']?server[""']?.*(?:/>|>)",
+							       RegexOptions.Compiled | RegexOptions.Singleline |
+							       RegexOptions.Multiline | RegexOptions.IgnoreCase |
+							       RegexOptions.CultureInvariant);
+		
+		static readonly Regex endOfTag = new Regex (@"</[\w:\.]+\s*?>",
+							    RegexOptions.Compiled | RegexOptions.Singleline |
+							    RegexOptions.Multiline | RegexOptions.IgnoreCase |
+							    RegexOptions.CultureInvariant);
+		
+		static readonly Regex expressionRegex = new Regex (@"<%.*?%>",
+								   RegexOptions.Compiled | RegexOptions.Singleline |
+								   RegexOptions.Multiline | RegexOptions.IgnoreCase |
+								   RegexOptions.CultureInvariant);
+
+		static readonly Regex clientCommentRegex = new Regex (@"<!--(.|\s)*?-->",
+								      RegexOptions.Compiled | RegexOptions.Multiline | RegexOptions.IgnoreCase |
+								      RegexOptions.CultureInvariant);
+		
 		ParserStack pstack;
 		BuilderLocationStack stack;
 		TemplateParser tparser;
@@ -188,24 +240,28 @@ namespace System.Web.Compilation
 		ILocation location;
 		bool isApplication;
 		StringBuilder tagInnerText = new StringBuilder ();
-		static Hashtable emptyHash = new Hashtable ();
+		static IDictionary emptyHash = new Dictionary <string, object> ();
 		bool inForm;
 		bool useOtherTags;
 		TagType lastTag;
+		AspComponentFoundry componentFoundry;
+		Stream inputStream;
 
+		public AspGenerator (TemplateParser tparser, AspComponentFoundry componentFoundry) : this (tparser)
+		{
+			this.componentFoundry = componentFoundry;
+		}
+		
 		public AspGenerator (TemplateParser tparser)
 		{
 			this.tparser = tparser;
 			text = new StringBuilder ();
 			stack = new BuilderLocationStack ();
-			rootBuilder = new RootBuilder (tparser);
-			stack.Push (rootBuilder, null);
-			tparser.RootBuilder = rootBuilder;
 			pstack = new ParserStack ();
 		}
 
 		public RootBuilder RootBuilder {
-			get { return tparser.RootBuilder; }
+			get { return rootBuilder; }
 		}
 
 		public AspParser Parser {
@@ -216,7 +272,6 @@ namespace System.Web.Compilation
 			get { return pstack.Filename; }
 		}
 
-#if NET_2_0
 		PageParserFilter PageParserFilter {
 			get {
 				if (tparser == null)
@@ -225,7 +280,236 @@ namespace System.Web.Compilation
 				return tparser.PageParserFilter;
 			}
 		}
-#endif
+
+		// KLUDGE WARNING
+		//
+		// The kludge to determine the base type of the to-be-generated ASP.NET class is
+		// very unfortunate but with our current parser it is, unfortunately, necessary. The
+		// reason for reading the entire file into memory and parsing it with a regexp is
+		// that we need to read the main directive (i.e. <%@Page %>, <%@Control %> etc),
+		// pass it to the page parser filter if it exists, and finally read the inherits
+		// attribute of the directive to get access to the base type of the class to be
+		// generated. On that type we check whether it is decorated with the
+		// FileLevelControlBuilder attribute and, if yes, use the indicated type as the
+		// RootBuilder. This is necessary for the ASP.NET MVC views using the "generic"
+		// inherits declaration to work properly. Our current parser is not able to parse
+		// the input file out of sequence (i.e. directives first, then the rest) so we need
+		// to do what we do below, alas.
+		IDictionary GetDirectiveAttributesDictionary (string skipKeyName, CaptureCollection names, CaptureCollection values)
+		{
+			var ret = new Dictionary <string, object> (StringComparer.OrdinalIgnoreCase);
+
+			int index = 0;
+			string keyName;
+			foreach (Capture c in names) {
+				keyName = c.Value;
+				if (String.Compare (skipKeyName, keyName, StringComparison.OrdinalIgnoreCase) == 0) {
+					index++;
+					continue;
+				}
+				
+				ret.Add (c.Value, values [index++].Value);
+			}
+
+			return ret;
+		}
+
+		string GetDirectiveName (CaptureCollection names)
+		{
+			string val;
+			foreach (Capture c in names) {
+				val = c.Value;
+				if (Directive.IsDirective (val))
+					return val;
+			}
+
+			return tparser.DefaultDirectiveName;
+		}
+
+		int GetLineNumberForIndex (string fileContents, int index)
+		{
+			int line = 1;
+			char c;
+			bool foundCR = false;
+			
+			for (int pos = 0; pos < index; pos++) {
+				c = fileContents [pos];
+				if (c == '\n' || foundCR) {
+					line++;
+					foundCR = false;
+				}
+				
+				foundCR = (c == '\r');
+			}
+
+			return line;
+		}
+
+		int GetNumberOfLinesForRange (string fileContents, int index, int length)
+		{
+			int lines = 0;
+			int stop = index + length;
+			char c;
+			bool foundCR = false;
+			
+			for (int pos = index; pos < stop; pos++) {
+				c = fileContents [pos];
+				if (c == '\n' || foundCR) {
+					lines++;
+					foundCR = false;
+				}
+
+				foundCR = (c == '\r');
+			}
+
+			return lines;
+		}
+		
+		Type GetInheritedType (string fileContents, string filename)
+		{
+			MatchCollection matches = DirectiveRegex.Matches (fileContents);
+			if (matches == null || matches.Count == 0)
+				return null;
+
+			string wantedDirectiveName = tparser.DefaultDirectiveName.ToLower (Helpers.InvariantCulture);
+			string directiveName;
+			GroupCollection groups;
+			CaptureCollection ccNames;
+			
+			foreach (Match match in matches) {
+				groups = match.Groups;
+				if (groups.Count < 6)
+					continue;
+
+				ccNames = groups [3].Captures;
+				directiveName = GetDirectiveName (ccNames);
+				if (String.IsNullOrEmpty (directiveName))
+					continue;
+				
+				if (String.Compare (directiveName.ToLower (Helpers.InvariantCulture), wantedDirectiveName, StringComparison.Ordinal) != 0)
+					continue;
+
+				var loc = new Location (null);
+				int index = match.Index;
+				
+				loc.Filename = filename;
+				loc.BeginLine = GetLineNumberForIndex (fileContents, index);
+				loc.EndLine = loc.BeginLine + GetNumberOfLinesForRange (fileContents, index, match.Length);
+				
+				tparser.Location = loc;
+				tparser.allowedMainDirectives = 2;
+				tparser.AddDirective (wantedDirectiveName, GetDirectiveAttributesDictionary (wantedDirectiveName, ccNames, groups [5].Captures));
+
+				return tparser.BaseType;
+			}
+			
+			return null;
+		}
+
+		string ReadFileContents (Stream inputStream, string filename)
+		{
+			string ret = null;
+			
+			if (inputStream != null) {
+				if (inputStream.CanSeek) {
+					long curPos = inputStream.Position;
+					inputStream.Seek (0, SeekOrigin.Begin);
+
+					Encoding enc = WebEncoding.FileEncoding;
+					StringBuilder sb = new StringBuilder ();
+					byte[] buffer = new byte [READ_BUFFER_SIZE];
+					int nbytes;
+					
+					while ((nbytes = inputStream.Read (buffer, 0, READ_BUFFER_SIZE)) > 0)
+						sb.Append (enc.GetString (buffer, 0, nbytes));
+					inputStream.Seek (curPos, SeekOrigin.Begin);
+					
+					ret = sb.ToString ();
+					sb.Length = 0;
+					sb.Capacity = 0;
+				} else {
+					FileStream fs = inputStream as FileStream;
+					if (fs != null) {
+						string fname = fs.Name;
+						try {
+							if (File.Exists (fname))
+								ret = File.ReadAllText (fname);
+						} catch {
+							// ignore
+						}
+					}
+				}
+			}
+
+			if (ret == null && !String.IsNullOrEmpty (filename) && String.Compare (filename, "@@inner_string@@", StringComparison.Ordinal) != 0) {
+				try {
+					if (File.Exists (filename))
+						ret = File.ReadAllText (filename);
+				} catch {
+					// ignore
+				}
+			}
+
+			return ret;
+		}
+		
+		Type GetRootBuilderType (Stream inputStream, string filename)
+		{
+			Type ret = null;
+			string fileContents;
+
+			if (tparser != null)
+				fileContents = ReadFileContents (inputStream, filename);
+			else
+				fileContents = null;
+			
+			if (!String.IsNullOrEmpty (fileContents)) {
+				Type inheritedType = GetInheritedType (fileContents, filename);
+				fileContents = null;
+				if (inheritedType != null) {
+					FileLevelControlBuilderAttribute attr;
+					
+					try {
+						object[] attrs = inheritedType.GetCustomAttributes (typeof (FileLevelControlBuilderAttribute), true);
+						if (attrs != null && attrs.Length > 0)
+							attr = attrs [0] as FileLevelControlBuilderAttribute;
+						else
+							attr = null;
+					} catch {
+						attr = null;
+					}
+
+					ret = attr != null ? attr.BuilderType : null;
+				}
+			}
+			
+			if (ret == null) {
+				if (tparser is PageParser)
+					return typeof (FileLevelPageControlBuilder);
+				else if (tparser is UserControlParser)
+					return typeof (FileLevelUserControlBuilder);
+				else
+					return typeof (RootBuilder);
+			} else
+				return ret;
+		}
+		
+		void CreateRootBuilder (Stream inputStream, string filename)
+		{
+			if (rootBuilder != null)
+				return;
+			
+			Type rootBuilderType = GetRootBuilderType (inputStream, filename);
+			rootBuilder = Activator.CreateInstance (rootBuilderType) as RootBuilder;
+			if (rootBuilder == null)
+				throw new HttpException ("Cannot create an instance of file-level control builder.");
+			rootBuilder.Init (tparser, null, null, null, null, null);
+			if (componentFoundry != null)
+				rootBuilder.Foundry = componentFoundry;
+			
+			stack.Push (rootBuilder, null);
+			tparser.RootBuilder = rootBuilder;
+		}
 		
 		BaseCompiler GetCompilerFromType ()
 		{
@@ -238,10 +522,9 @@ namespace System.Web.Compilation
 
 			if (type == typeof (UserControlParser))
 				return new UserControlCompiler ((UserControlParser) tparser);
-#if NET_2_0
+
 			if (type == typeof(MasterPageParser))
 				return new MasterPageCompiler ((MasterPageParser) tparser);
-#endif
 
 			throw new Exception ("Got type: " + type);
 		}
@@ -252,10 +535,10 @@ namespace System.Web.Compilation
 			parser.Error += new ParseErrorHandler (ParseError);
 			parser.TagParsed += new TagParsedHandler (TagParsed);
 			parser.TextParsed += new TextParsedHandler (TextParsed);
-#if NET_2_0
 			parser.ParsingComplete += new ParsingCompleteHandler (ParsingCompleted);
 			tparser.AspGenerator = this;
-#endif
+			CreateRootBuilder (inputStream, filename);
+
 			if (!pstack.Push (parser))
 				throw new ParseException (Location, "Infinite recursion detected including file: " + filename);
 
@@ -266,26 +549,59 @@ namespace System.Web.Compilation
 				
 				tparser.AddDependency (arvp);
 			}
-#if NET_2_0
-			tparser.MD5Checksum = parser.MD5Checksum;
-#endif
 		}
 		
-#if NET_2_0
 		void InitParser (string filename)
 		{
 			StreamReader reader = new StreamReader (filename, WebEncoding.FileEncoding);
 			InitParser (reader, filename);
 		}
-#endif
+
+		void CheckForDuplicateIds (ControlBuilder root, Stack scopes)
+		{
+			if (root == null)
+				return;
+			
+			if (scopes == null)
+				scopes = new Stack ();			
+
+			Dictionary <string, bool> ids;
+			
+			if (scopes.Count == 0 || root.IsNamingContainer) {
+				ids = new Dictionary <string, bool> (StringComparer.Ordinal);
+				scopes.Push (ids);
+			} else {
+				ids = scopes.Peek () as Dictionary <string, bool>;
+			}
+			
+			if (ids == null)
+				return;
+
+			ControlBuilder cb;
+			string id;
+			ArrayList children = root.Children;
+			if (children != null) {
+				foreach (object o in children) {
+					cb = o as ControlBuilder;
+					if (cb == null)
+						continue;
+
+					id = cb.ID;
+					if (id == null || id.Length == 0)
+						continue;
+				
+					if (ids.ContainsKey (id))
+						throw new ParseException (cb.Location, "Id '" + id + "' is already used by another control.");
+
+					ids.Add (id, true);
+					CheckForDuplicateIds (cb, scopes);
+				}
+			}
+		}
 		
 		public void Parse (string file)
 		{
-#if ONLY_1_1
-			Parse (file, true);
-#else
 			Parse (file, false);
-#endif
 		}
 		
 		public void Parse (TextReader reader, string filename, bool doInitParser)
@@ -300,15 +616,18 @@ namespace System.Web.Compilation
 				if (text.Length > 0)
 					FlushText ();
 
+				tparser.MD5Checksum = pstack.Parser.MD5Checksum;
 				pstack.Pop ();
 
 #if DEBUG
-				PrintTree (rootBuilder, 0);
+				PrintTree (RootBuilder, 0);
 #endif
 
 				if (stack.Count > 1 && pstack.Count == 0)
 					throw new ParseException (stack.Builder.Location,
 								  "Expecting </" + stack.Builder.TagName + "> " + stack.Builder);
+
+				CheckForDuplicateIds (RootBuilder, null);
 			} finally {
 				if (reader != null)
 					reader.Close ();
@@ -317,6 +636,7 @@ namespace System.Web.Compilation
 
 		public void Parse (Stream stream, string filename, bool doInitParser)
 		{
+			inputStream = stream;
 			Parse (new StreamReader (stream, WebEncoding.FileEncoding), filename, doInitParser);
 		}
 		
@@ -328,7 +648,6 @@ namespace System.Web.Compilation
 
 		public void Parse ()
 		{
-#if NET_2_0
 			string inputFile = tparser.InputFile;
 			TextReader inputReader = tparser.Reader;
 
@@ -358,18 +677,15 @@ namespace System.Web.Compilation
 				if (inputReader != null)
 					inputReader.Close ();
 			}
-#else
-			Parse (Path.GetFullPath (tparser.InputFile));
-#endif
 		}
 
-		internal static void AddTypeToCache (ArrayList dependencies, string inputFile, Type type)
+		internal static void AddTypeToCache (List <string> dependencies, string inputFile, Type type)
 		{
 			if (type == null || inputFile == null || inputFile.Length == 0)
 				return;
 
 			if (dependencies != null && dependencies.Count > 0) {
-				string [] deps = (string []) dependencies.ToArray (typeof (string));
+				string [] deps = dependencies.ToArray ();
 				HttpContext ctx = HttpContext.Current;
 				HttpRequest req = ctx != null ? ctx.Request : null;
 				
@@ -392,7 +708,6 @@ namespace System.Web.Compilation
 			}
 
 			Parse ();
-
 			BaseCompiler compiler = GetCompilerFromType ();
 			
 			type = compiler.GetCompiledType ();
@@ -408,8 +723,8 @@ namespace System.Web.Compilation
 
 			string i = new string ('\t', indent);
 			Console.Write (i);
-			Console.WriteLine ("b: {0} id: {1} type: {2} parent: {3}",
-					   builder, builder.ID, builder.ControlType, builder.ParentBuilder);
+			Console.WriteLine ("b: {0}; naming container: {1}; id: {2}; type: {3}; parent: {4}",
+					   builder, builder.IsNamingContainer, builder.ID, builder.ControlType, builder.ParentBuilder);
 
 			if (builder.Children != null)
 			foreach (object o in builder.Children) {
@@ -446,9 +761,6 @@ namespace System.Web.Compilation
 		// The kludge supports only self-closing tags inside attributes.
 		//
 		// KLUDGE WARNING!!
-		static readonly Regex runatServer=new Regex (@"<[\w:\.]+.*?runat=[""']?server[""']?.*?/>",
-							     RegexOptions.Compiled | RegexOptions.Singleline |
-							     RegexOptions.Multiline | RegexOptions.IgnoreCase);
 		bool ProcessTagsInAttributes (ILocation location, string tagid, TagAttributes attributes, TagType type)
 		{
 			if (attributes == null || attributes.Count == 0)
@@ -485,8 +797,8 @@ namespace System.Web.Compilation
 				length = group.Length;
 
 				TextParsed (location, String.Format (" {0}=\"{1}", key, index > 0 ? value.Substring (0, index) : String.Empty));;
-				FlushText ();				
-				ParseAttributeTag (group.Value);
+				FlushText ();
+				ParseAttributeTag (group.Value, location);
 				if (index + length < value.Length)
 					TextParsed (location, value.Substring (index + length) + "\"");
 				else
@@ -503,9 +815,11 @@ namespace System.Web.Compilation
 			return retval;
 		}
 
-		void ParseAttributeTag (string code)
+		void ParseAttributeTag (string code, ILocation location)
 		{
-			AspParser parser = new AspParser ("@@attribute_tag@@", new StringReader (code));
+			AspParser outerParser = location as AspParser;
+			int positionOffset = outerParser != null ? outerParser.BeginPosition : 0;
+			AspParser parser = new AspParser ("@@attribute_tag@@", new StringReader (code), location.BeginLine - 1, positionOffset, outerParser);
 			parser.Error += new ParseErrorHandler (ParseError);
 			parser.TagParsed += new TagParsedHandler (TagParsed);
 			parser.TextParsed += new TextParsedHandler (TextParsed);
@@ -514,32 +828,75 @@ namespace System.Web.Compilation
 				FlushText ();
 		}
 
-#if NET_2_0
 		void ParsingCompleted ()
 		{
 			PageParserFilter pfilter = PageParserFilter;
 			if (pfilter == null)
 				return;
 
-			pfilter.ParseComplete (rootBuilder);
+			pfilter.ParseComplete (RootBuilder);
+		}
+
+		void CheckIfIncludeFileIsSecure (string filePath)
+		{
+			if (filePath == null || filePath.Length == 0)
+				return;
+			
+			// a bit slow, but fully portable
+			string newdir = null;
+			Exception exception = null;
+			try {
+				string origdir = Directory.GetCurrentDirectory ();
+				Directory.SetCurrentDirectory (Path.GetDirectoryName (filePath));
+				newdir = Directory.GetCurrentDirectory ();
+				Directory.SetCurrentDirectory (origdir);
+				if (newdir [newdir.Length - 1] != '/')
+					newdir += "/";
+			} catch (DirectoryNotFoundException) {
+				return; // will be converted into 404
+			} catch (FileNotFoundException) {
+				return; // as above
+			} catch (Exception ex) {
+				// better safe than sorry
+				exception = ex;
+			}
+
+			if (exception != null || !StrUtils.StartsWith (newdir, HttpRuntime.AppDomainAppPath))
+				throw new ParseException (Location, "Files above the application's root directory cannot be included.");
+		}
+
+		string ChopOffTagStart (ILocation location, string content, string tagid)
+		{
+			string tagstart = '<' + tagid;
+			if (content.StartsWith (tagstart)) {
+				TextParsed (location, tagstart);
+				content = content.Substring (tagstart.Length);
+			}
+
+			return content;
 		}
-#endif
 		
 		void TagParsed (ILocation location, TagType tagtype, string tagid, TagAttributes attributes)
 		{
+			bool tagIgnored;
+			
 			this.location = new Location (location);
 			if (tparser != null)
 				tparser.Location = location;
 
-			if (text.Length != 0)
-				FlushText (lastTag == TagType.CodeRender);
-
-			if (0 == String.Compare (tagid, "script", true, CultureInfo.InvariantCulture)) {
+			if (text.Length != 0) {
+				bool ignoreEmptyString = lastTag == TagType.CodeRender;
+				FlushText (ignoreEmptyString);
+			}
+			
+			if (0 == String.Compare (tagid, "script", true, Helpers.InvariantCulture)) {
 				bool in_script = (inScript || ignore_text);
-				if (in_script || (tagtype != TagType.Close && attributes != null)) {
-					if ((in_script || attributes.IsRunAtServer ()) && ProcessScript (tagtype, attributes))
+				if (in_script) {
+					if (ProcessScript (tagtype, attributes))
+						return;
+				} else
+					if (ProcessScript (tagtype, attributes))
 						return;
-				}
 			}
 
 			lastTag = tagtype;
@@ -551,8 +908,9 @@ namespace System.Web.Compilation
 				tparser.AddDirective (tagid, attributes.GetDictionary (null));
 				break;
 			case TagType.Tag:
-				if (ProcessTag (location, tagid, attributes, tagtype)) {
-					useOtherTags = true;
+				if (ProcessTag (location, tagid, attributes, tagtype, out tagIgnored)) {
+					if (!tagIgnored)
+						useOtherTags = true;
 					break;
 				}
 
@@ -564,7 +922,7 @@ namespace System.Web.Compilation
 				{
 					string plainText = location.PlainText;
 					if (!ProcessTagsInAttributes (location, tagid, attributes, TagType.Tag))
-						TextParsed (location, plainText);
+						TextParsed (location, ChopOffTagStart (location, plainText, tagid));
 				}
 				break;
 			case TagType.Close:
@@ -576,19 +934,20 @@ namespace System.Web.Compilation
 				break;
 			case TagType.SelfClosing:
 				int count = stack.Count;
-				if (!ProcessTag (location, tagid, attributes, tagtype)) {
+				if (!ProcessTag (location, tagid, attributes, tagtype, out tagIgnored) && !tagIgnored) {
 					string plainText = location.PlainText;
 					if (!ProcessTagsInAttributes (location, tagid, attributes, TagType.SelfClosing))
-						TextParsed (location, plainText);
+						TextParsed (location, ChopOffTagStart (location, plainText, tagid));
 				} else if (stack.Count != count) {
 					CloseControl (tagid);
 				}
 				break;
 			case TagType.DataBinding:
-				goto case TagType.CodeRender;
 			case TagType.CodeRenderExpression:
-				goto case TagType.CodeRender;
 			case TagType.CodeRender:
+#if NET_4_0
+			case TagType.CodeRenderEncode:
+#endif
 				if (isApplication)
 					throw new ParseException (location, "Invalid content for application file.");
 			
@@ -605,7 +964,6 @@ namespace System.Web.Compilation
 
 				if (isvirtual) {
 					bool parsed = false;
-#if NET_2_0
 					VirtualPathProvider vpp = HostingEnvironment.VirtualPathProvider;
 
 					if (vpp.FileExists (file)) {
@@ -615,12 +973,12 @@ namespace System.Web.Compilation
 							parsed = true;
 						}
 					}
-#endif
 					
 					if (!parsed)
 						Parse (tparser.MapPath (file), true);
 				} else {
 					string includeFilePath = GetIncludeFilePath (tparser.ParserDir, file);
+					CheckIfIncludeFileIsSecure (includeFilePath);
 					tparser.PushIncludeDir (Path.GetDirectoryName (includeFilePath));
 					try {
 						Parse (includeFilePath, true);
@@ -643,7 +1001,7 @@ namespace System.Web.Compilation
 
 			for (int idx = otags.Count - 1; idx >= 0; idx--) {
 				string otagid = (string) otags [idx];
-				if (0 == String.Compare (tagid, otagid, true, CultureInfo.InvariantCulture)) {
+				if (0 == String.Compare (tagid, otagid, true, Helpers.InvariantCulture)) {
 					do {
 						otags.RemoveAt (idx);
 					} while (otags.Count - 1 >= idx);
@@ -660,22 +1018,122 @@ namespace System.Web.Compilation
 
 			return Path.GetFullPath (Path.Combine (basedir, filename));
 		}
+
+		delegate bool CheckBlockEnd (string text);
 		
+		bool CheckTagEndNeeded (string text)
+		{
+			return !text.EndsWith ("/>");
+		}
+		
+		List <TextBlock> FindRegexBlocks (Regex rxStart, Regex rxEnd, CheckBlockEnd checkEnd, IList blocks, TextBlockType typeForMatches, bool discardBlocks)
+		{
+			var ret = new List <TextBlock> ();
+			foreach (TextBlock block in blocks) {
+				if (block.Type != TextBlockType.Verbatim) {
+					ret.Add (block);
+					continue;
+				}
+
+				int lastIndex = 0, index;
+				MatchCollection matches = rxStart.Matches (block.Content);
+				bool foundMatches = matches.Count > 0;
+				foreach (Match match in matches) {
+					foundMatches = true;
+					index = match.Index;
+					if (lastIndex < index)
+						ret.Add (new TextBlock (TextBlockType.Verbatim, block.Content.Substring (lastIndex, index - lastIndex)));
+
+					string value = match.Value;
+					if (rxEnd != null && checkEnd (value)) {
+						int startFrom = index + value.Length;
+						Match m = rxEnd.Match (block.Content, startFrom);
+						if (m.Success)
+							value += block.Content.Substring (startFrom, m.Index - startFrom) + m.Value;
+					}
+
+					if (!discardBlocks)
+						ret.Add (new TextBlock (typeForMatches, value));
+					lastIndex = index + value.Length;
+				}
+
+				if (lastIndex > 0 && lastIndex < block.Content.Length)
+					ret.Add (new TextBlock (TextBlockType.Verbatim, block.Content.Substring (lastIndex)));
+
+				if (!foundMatches)
+					ret.Add (block);
+			}
+
+			return ret;
+		}
+		
+		IList SplitTextIntoBlocks (string text)
+		{
+			var ret = new List <TextBlock> ();
+
+			ret.Add (new TextBlock (TextBlockType.Verbatim, text));
+			ret = FindRegexBlocks (clientCommentRegex, null, null, ret, TextBlockType.Comment, false);
+			ret = FindRegexBlocks (runatServer, endOfTag, CheckTagEndNeeded, ret, TextBlockType.Tag, false);
+			ret = FindRegexBlocks (expressionRegex, null, null, ret, TextBlockType.Expression, false);
+
+			return ret;
+		}
+
 		void TextParsed (ILocation location, string text)
 		{
 			if (ignore_text)
 				return;
 
-			if (text.IndexOf ("<%") != -1 && !inScript) {
-				if (this.text.Length > 0)
-					FlushText (true);
-				CodeRenderParser r = new CodeRenderParser (text, stack.Builder);
-				r.AddChildren (this);
+			if (inScript) {
+				this.text.Append (text);
+				FlushText (true);
 				return;
 			}
-			
-			this.text.Append (text);
-			//PrintLocation (location);
+
+			IList blocks = SplitTextIntoBlocks (text);
+			foreach (TextBlock block in blocks) {
+				switch (block.Type) {
+					case TextBlockType.Verbatim:
+						this.text.Append (block.Content);
+						break;
+
+					case TextBlockType.Expression:
+						if (this.text.Length > 0)
+							FlushText (true);
+						CodeRenderParser r = new CodeRenderParser (block.Content, stack.Builder, location);
+						r.AddChildren (this);
+						break;
+
+					case TextBlockType.Tag:
+						ParseAttributeTag (block.Content, location);
+						break;
+
+					case TextBlockType.Comment: {
+						this.text.Append ("<!--");
+						FlushText (true);
+						string blockToParse = block.Content.Substring (4, block.Length - 7);
+						bool condEndif;
+						if (blockToParse.EndsWith ("<![endif]")) {
+							blockToParse = blockToParse.Substring (0, blockToParse.Length - 9);
+							condEndif = true;
+						} else
+							condEndif = false;
+
+						AspParser outerParser = location as AspParser;
+						int positionOffset = outerParser != null ? outerParser.BeginPosition : 0;
+						AspParser parser = new AspParser ("@@comment_code@@", new StringReader (blockToParse), location.BeginLine - 1, positionOffset, outerParser);
+						parser.Error += new ParseErrorHandler (ParseError);
+						parser.TagParsed += new TagParsedHandler (TagParsed);
+						parser.TextParsed += new TextParsedHandler (TextParsed);
+						parser.Parse ();
+						if (condEndif)
+							this.text.Append ("<![endif]");
+						this.text.Append ("-->");
+						FlushText (true);
+						break;
+					}
+				}
+			}
 		}
 
 		void FlushText ()
@@ -692,11 +1150,10 @@ namespace System.Web.Compilation
 				return;
 			
 			if (inScript) {
-#if NET_2_0
 				PageParserFilter pfilter = PageParserFilter;
 				if (pfilter != null && !pfilter.ProcessCodeConstruct (CodeConstructType.ScriptTag, t))
 					return;
-#endif
+
 				tparser.Scripts.Add (new ServerSideScript (t, new System.Web.Compilation.Location (tparser.Location)));
 				return;
 			}
@@ -711,7 +1168,6 @@ namespace System.Web.Compilation
 			}
 		}
 
-#if NET_2_0
 		bool BuilderHasOtherThan (Type type, ControlBuilder cb)
 		{
 			ArrayList al = cb.OtherTags;
@@ -754,12 +1210,11 @@ namespace System.Web.Compilation
 			if (!typeof (System.Web.UI.WebControls.Content).IsAssignableFrom (cb.ControlType))
 				return true;
 
-			if (BuilderHasOtherThan (typeof (System.Web.UI.WebControls.Content), rootBuilder))
+			if (BuilderHasOtherThan (typeof (System.Web.UI.WebControls.Content), RootBuilder))
 				return false;
 			
 			return true;
 		}
-#endif
 
 		public void AddControl (Type type, IDictionary attributes)
 		{
@@ -771,16 +1226,24 @@ namespace System.Web.Compilation
 				parent.AppendSubBuilder (builder);
 		}
 		
-		bool ProcessTag (ILocation location, string tagid, TagAttributes atts, TagType tagtype)
+		bool ProcessTag (ILocation location, string tagid, TagAttributes atts, TagType tagtype, out bool ignored)
 		{
+			ignored = false;
 			if (isApplication) {
-				if (String.Compare (tagid, "object", true, CultureInfo.InvariantCulture) != 0)
+				if (String.Compare (tagid, "object", true, Helpers.InvariantCulture) != 0)
 					throw new ParseException (location, "Invalid tag for application file.");
 			}
 
 			ControlBuilder parent = stack.Builder;
 			ControlBuilder builder = null;
-			Hashtable htable = (atts != null) ? atts.GetDictionary (null) : emptyHash;
+			if (parent != null && parent.ControlType == typeof (HtmlTable) &&
+			    (String.Compare (tagid, "thead", true, Helpers.InvariantCulture) == 0 ||
+			     String.Compare (tagid, "tbody", true, Helpers.InvariantCulture) == 0)) {
+				ignored = true;
+				return true;
+			}
+				
+			IDictionary htable = (atts != null) ? atts.GetDictionary (null) : emptyHash;
 			if (stack.Count > 1) {
 				try {
 					builder = parent.CreateSubBuilder (tagid, htable, null, tparser, location);
@@ -791,15 +1254,22 @@ namespace System.Web.Compilation
 				}
 			}
 
-			if (builder == null && atts != null && atts.IsRunAtServer ()) {
+			bool runatServer = atts != null && atts.IsRunAtServer ();
+			if (builder == null && runatServer) {
 				string id = htable ["id"] as string;
 				if (id != null && !CodeGenerator.IsValidLanguageIndependentIdentifier (id))
 					throw new ParseException (Location, "'" + id + "' is not a valid identifier");
 					
 				try {
-					builder = rootBuilder.CreateSubBuilder (tagid, htable, null, tparser, location);
+					builder = RootBuilder.CreateSubBuilder (tagid, htable, null, tparser, location);
 				} catch (TypeLoadException e) {
 					throw new ParseException (Location, "Type not found.", e);
+				} catch (HttpException e) {
+					CompilationException inner = e.InnerException as CompilationException;
+					if (inner != null)
+						throw inner;
+					
+					throw new ParseException (Location, e.Message, e);
 				} catch (Exception e) {
 					throw new ParseException (Location, e.Message, e);
 				}
@@ -808,14 +1278,18 @@ namespace System.Web.Compilation
 			if (builder == null)
 				return false;
 
-#if NET_2_0
+			// This is as good as we can do for now - if the parsed location contains
+			// both expressions and code render blocks then we're out of luck...
+			string plainText = location.PlainText;
+			if (!runatServer && plainText.IndexOf ("<%$") == -1&& plainText.IndexOf ("<%") > -1)
+				return false;
+
 			PageParserFilter pfilter = PageParserFilter;
 			if (pfilter != null && !pfilter.AllowControl (builder.ControlType, builder))
 				throw new ParseException (Location, "Control type '" + builder.ControlType + "' not allowed.");
 			
 			if (!OtherControlsAllowed (builder))
 				throw new ParseException (Location, "Only Content controls are allowed directly in a content page that contains Content controls.");
-#endif
 			
 			builder.Location = location;
 			builder.ID = htable ["id"] as string;
@@ -834,7 +1308,7 @@ namespace System.Web.Compilation
 			} else {
 				if (!isApplication && builder is ObjectTagBuilder) {
 					ObjectTagBuilder ot = (ObjectTagBuilder) builder;
-					if (ot.Scope != null && ot.Scope != "")
+					if (ot.Scope != null && ot.Scope.Length > 0)
 						throw new ParseException (location, "Scope not allowed here");
 
 					if (tagtype == TagType.Tag) {
@@ -869,7 +1343,7 @@ namespace System.Web.Compilation
 					CheckLanguage (language);
 					string src = (string) attributes ["src"];
 					if (src != null) {
-						if (src == "")
+						if (src.Length == 0)
 							throw new ParseException (Parser,
 								"src cannot be an empty string");
 
@@ -893,7 +1367,14 @@ namespace System.Web.Compilation
 						Parser.VerbatimID = "script";
 						javascript = true;
 					}
-					TextParsed (location, location.PlainText);
+					string content = location.PlainText;
+					/* HACK, HACK, HACK */
+					if (content.StartsWith ("<script")) {
+						TextParsed (location, "<script");
+						content = content.Substring (7);
+					}
+
+					TextParsed (location, content);
 					return true;
 				}
 			}
@@ -918,8 +1399,8 @@ namespace System.Web.Compilation
 		{
 			ControlBuilder current = stack.Builder;
 			string btag = current.OriginalTagName;
-			if (String.Compare (btag, "tbody", true, CultureInfo.InvariantCulture) != 0 &&
-			    String.Compare (tagid, "tbody", true, CultureInfo.InvariantCulture) == 0) {
+			if (String.Compare (btag, "tbody", true, Helpers.InvariantCulture) != 0 &&
+			    String.Compare (tagid, "tbody", true, Helpers.InvariantCulture) == 0) {
 				if (!current.ChildrenAsProperties) {
 					try {
 						TextParsed (location, location.PlainText);
@@ -928,8 +1409,11 @@ namespace System.Web.Compilation
 				}
 				return true;
 			}
+
+			if (current.ControlType == typeof (HtmlTable) && String.Compare (tagid, "thead", true, Helpers.InvariantCulture) == 0)
+				return true;
 			
-			if (0 != String.Compare (tagid, btag, true, CultureInfo.InvariantCulture))
+			if (0 != String.Compare (tagid, btag, true, Helpers.InvariantCulture))
 				return false;
 
 			// if (current is TemplateBuilder)
@@ -954,17 +1438,19 @@ namespace System.Web.Compilation
 			return true;
 		}
 
-#if NET_2_0
 		CodeConstructType MapTagTypeToConstructType (TagType tagtype)
 		{
 			switch (tagtype) {
-				case TagType.DataBinding:
+				case TagType.CodeRenderExpression:
 					return CodeConstructType.ExpressionSnippet;
 
 				case TagType.CodeRender:
+#if NET_4_0
+				case TagType.CodeRenderEncode:
+#endif
 					return CodeConstructType.CodeSnippet;
 
-				case TagType.CodeRenderExpression:
+				case TagType.DataBinding:
 					return CodeConstructType.DataBindingSnippet;
 
 				default:
@@ -972,14 +1458,21 @@ namespace System.Web.Compilation
 			}
 		}
 		
-#endif
 		bool ProcessCode (TagType tagtype, string code, ILocation location)
 		{
-#if NET_2_0
 			PageParserFilter pfilter = PageParserFilter;
-			if (pfilter != null && (!pfilter.AllowCode || !pfilter.ProcessCodeConstruct (MapTagTypeToConstructType (tagtype), code)))
+			// LAMESPEC:
+			//
+			// http://msdn.microsoft.com/en-us/library/system.web.ui.pageparserfilter.processcodeconstruct.aspx
+			//
+			// The above page says if false is returned then we should NOT process the
+			// code further, wheras in reality it's the other way around. The
+			// ProcessCodeConstruct return value means whether or not the filter
+			// _processed_ the code.
+			//
+			if (pfilter != null && (!pfilter.AllowCode || pfilter.ProcessCodeConstruct (MapTagTypeToConstructType (tagtype), code)))
 				return true;
-#endif
+
 			ControlBuilder b = null;
 			if (tagtype == TagType.CodeRender)
 				b = new CodeRenderBuilder (code, false, location);
@@ -987,6 +1480,10 @@ namespace System.Web.Compilation
 				b = new CodeRenderBuilder (code, true, location);
 			else if (tagtype == TagType.DataBinding)
 				b = new DataBindingBuilder (code, location);
+#if NET_4_0
+			else if (tagtype == TagType.CodeRenderEncode)
+				b = new CodeRenderBuilder (code, true, location, true);
+#endif
 			else
 				throw new HttpException ("Should never happen");
 
@@ -1003,16 +1500,11 @@ namespace System.Web.Compilation
 			if (lang == null || lang == "")
 				return;
 
-			if (String.Compare (lang, tparser.Language, true, CultureInfo.InvariantCulture) == 0)
+			if (String.Compare (lang, tparser.Language, true, Helpers.InvariantCulture) == 0)
 				return;
 
-#if NET_2_0
 			CompilationSection section = (CompilationSection) WebConfigurationManager.GetWebApplicationSection ("system.web/compilation");
 			if (section.Compilers[tparser.Language] != section.Compilers[lang])
-#else
-			CompilationConfiguration cfg = CompilationConfiguration.GetInstance (HttpContext.Current); 
-			if (!cfg.Compilers.CompareLanguages (tparser.Language, lang))
-#endif
 				throw new ParseException (Location,
 						String.Format ("Trying to mix language '{0}' and '{1}'.", 
 								tparser.Language, lang));
@@ -1024,23 +1516,62 @@ namespace System.Web.Compilation
 			string str;
 			ControlBuilder builder;
 			AspGenerator generator;
+			ILocation location;
 			
-			public CodeRenderParser (string str, ControlBuilder builder)
+			public CodeRenderParser (string str, ControlBuilder builder, ILocation location)
 			{
 				this.str = str;
 				this.builder = builder;
+				this.location = location;
 			}
 
 			public void AddChildren (AspGenerator generator)
 			{
 				this.generator = generator;
 				int index = str.IndexOf ("<%");
-				if (index > 0) {
-					TextParsed (null, str.Substring (0, index));
-					str = str.Substring (index);
-				}
+				if (index > 0)
+					DoParseExpressions (str);
+				else
+					DoParse (str);
+			}
 
-				AspParser parser = new AspParser ("@@nested_tag@@", new StringReader (str));
+			void DoParseExpressions (string str)
+			{
+				int startIndex = 0, index = 0;
+				Regex codeDirective = new Regex ("(<%(?!@)(?<code>(.|\\s)*?)%>)|(<[\\w:\\.]+.*?runat=[\"']?server[\"']?.*?/>)",
+								 RegexOptions.Multiline | RegexOptions.Compiled | RegexOptions.CultureInvariant);
+				Match match;
+				int strLen = str.Length;
+				
+				while (index > -1 && startIndex < strLen) {
+					match = codeDirective.Match (str, index);
+					
+					if (match.Success) {
+						string value = match.Value;
+						index = match.Index;
+						if (index > startIndex)
+							TextParsed (null, str.Substring (startIndex, index - startIndex));
+						DoParse (value);
+						index += value.Length;
+						startIndex = index;
+					} else
+						break;
+
+					if (index < strLen)
+						index = str.IndexOf ('<', index);
+					else
+						break;
+				}
+				
+				if (startIndex < strLen)
+					TextParsed (null, str.Substring (startIndex));
+			}
+			
+			void DoParse (string str)
+			{
+				AspParser outerParser = location as AspParser;
+				int positionOffset = outerParser != null ? outerParser.BeginPosition : 0;
+				AspParser parser = new AspParser ("@@code_render@@", new StringReader (str), location.BeginLine - 1, positionOffset, outerParser);
 				parser.Error += new ParseErrorHandler (ParseError);
 				parser.TagParsed += new TagParsedHandler (TagParsed);
 				parser.TextParsed += new TextParsedHandler (TextParsed);
@@ -1057,13 +1588,18 @@ namespace System.Web.Compilation
 					case TagType.CodeRenderExpression:
 						builder.AppendSubBuilder (new CodeRenderBuilder (tagid, true, location));
 						break;
-						
+#if NET_4_0
+					case TagType.CodeRenderEncode:
+						builder.AppendSubBuilder (new CodeRenderBuilder (tagid, true, location, true));
+						break;
+#endif
 					case TagType.DataBinding:
 						builder.AppendSubBuilder (new DataBindingBuilder (tagid, location));
 						break;
 
 					case TagType.Tag:
 					case TagType.SelfClosing:
+					case TagType.Close:
 						if (generator != null)
 							generator.TagParsed (location, tagtype, tagid, attributes);
 						else
@@ -1071,7 +1607,9 @@ namespace System.Web.Compilation
 						break;
 						
 					default:
-						builder.AppendLiteralString (location.PlainText);
+						string text = location.PlainText;
+						if (text != null && text.Trim ().Length > 0)
+							builder.AppendLiteralString (text);
 						break;
 				}
 			}
