@@ -4,7 +4,7 @@
 // Author:
 //	Sebastien Pouliot  <sebastien@ximian.com>
 //
-// Copyright (C) 2005 Novell, Inc (http://www.novell.com)
+// Copyright (C) 2005-2010 Novell, Inc (http://www.novell.com)
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the
@@ -32,73 +32,55 @@ using System.Globalization;
 using System.Security.Permissions;
 using System.Web.Util;
 
-namespace System.Web.UI.WebControls {
-
+namespace System.Web.UI.WebControls
+{
 	// CAS
 	[AspNetHostingPermission (SecurityAction.LinkDemand, Level = AspNetHostingPermissionLevel.Minimal)]
 	[AspNetHostingPermission (SecurityAction.InheritanceDemand, Level = AspNetHostingPermissionLevel.Minimal)]
 	// attributes
 	[Designer ("System.Web.UI.Design.WebControls.DataListDesigner, " + Consts.AssemblySystem_Design, "System.ComponentModel.Design.IDesigner")]
-#if NET_2_0
 	[ControlValueProperty ("SelectedValue")]
 	[Editor ("System.Web.UI.Design.WebControls.DataListComponentEditor, " + Consts.AssemblySystem_Design, "System.ComponentModel.ComponentEditor, " + Consts.AssemblySystem)]
-#else
-	[Editor ("System.Web.UI.Design.WebControls.DataListComponentEditor, " + Consts.AssemblySystem_Design, typeof (System.ComponentModel.ComponentEditor))]
-#endif
-	public class DataList : BaseDataList, INamingContainer, IRepeatInfoUser {
-
+	public class DataList : BaseDataList, INamingContainer, IRepeatInfoUser
+	{
 		public const string CancelCommandName = "Cancel";
 		public const string DeleteCommandName = "Delete";
 		public const string EditCommandName = "Edit";
 		public const string SelectCommandName = "Select";
 		public const string UpdateCommandName = "Update";
 
+		static readonly object cancelCommandEvent = new object ();
+		static readonly object deleteCommandEvent = new object ();
+		static readonly object editCommandEvent = new object ();
+		static readonly object itemCommandEvent = new object ();
+		static readonly object itemCreatedEvent = new object ();
+		static readonly object itemDataBoundEvent = new object ();
+		static readonly object updateCommandEvent = new object ();
 
-		private static readonly object cancelCommandEvent = new object ();
-		private static readonly object deleteCommandEvent = new object ();
-		private static readonly object editCommandEvent = new object ();
-		private static readonly object itemCommandEvent = new object ();
-		private static readonly object itemCreatedEvent = new object ();
-		private static readonly object itemDataBoundEvent = new object ();
-		private static readonly object updateCommandEvent = new object ();
+		TableItemStyle alternatingItemStyle;
+		TableItemStyle editItemStyle;
+		TableItemStyle footerStyle;
+		TableItemStyle headerStyle;
+		TableItemStyle itemStyle;
+		TableItemStyle selectedItemStyle;
+		TableItemStyle separatorStyle;
 
-		private TableItemStyle alternatingItemStyle;
-		private TableItemStyle editItemStyle;
-		private TableItemStyle footerStyle;
-		private TableItemStyle headerStyle;
-		private TableItemStyle itemStyle;
-		private TableItemStyle selectedItemStyle;
-		private TableItemStyle separatorStyle;
+		ITemplate alternatingItemTemplate;
+		ITemplate editItemTemplate;
+		ITemplate footerTemplate;
+		ITemplate headerTemplate;
+		ITemplate itemTemplate;
+		ITemplate selectedItemTemplate;
+		ITemplate separatorTemplate;
 
-		private ITemplate alternatingItemTemplate;
-		private ITemplate editItemTemplate;
-		private ITemplate footerTemplate;
-		private ITemplate headerTemplate;
-		private ITemplate itemTemplate;
-		private ITemplate selectedItemTemplate;
-		private ITemplate separatorTemplate;
+		DataListItemCollection items;
+		ArrayList list;
+		int idx;
 
-		private DataListItemCollection items;
-		private ArrayList list;
-		private int idx;
-
-#if NET_2_0
-		//private int editIndex;
-		private int selectedIndex;
-
-		public DataList ()
-		{
-			//editIndex = -1;
-			selectedIndex = -1;
-			idx = -1;
-		}
-#else
 		public DataList ()
 		{
 			idx = -1;
 		}
-#endif
-
 
 		[DefaultValue (null)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Content)]
@@ -110,9 +92,8 @@ namespace System.Web.UI.WebControls {
 			get {
 				if (alternatingItemStyle == null) {
 					alternatingItemStyle = new TableItemStyle ();
-					if (IsTrackingViewState) {
+					if (IsTrackingViewState)
 						alternatingItemStyle.TrackViewState ();
-					}
 				}
 				return alternatingItemStyle;
 			}
@@ -154,9 +135,8 @@ namespace System.Web.UI.WebControls {
 			get {
 				if (editItemStyle == null) {
 					editItemStyle = new TableItemStyle ();
-					if (IsTrackingViewState) {
+					if (IsTrackingViewState)
 						editItemStyle.TrackViewState ();
-					}
 				}
 				return editItemStyle;
 			}
@@ -194,9 +174,8 @@ namespace System.Web.UI.WebControls {
 			get {
 				if (footerStyle == null) {
 					footerStyle = new TableItemStyle ();
-					if (IsTrackingViewState) {
+					if (IsTrackingViewState)
 						footerStyle.TrackViewState ();
-					}
 				}
 				return footerStyle;
 			}
@@ -217,13 +196,8 @@ namespace System.Web.UI.WebControls {
 		[DefaultValue (GridLines.None)]
 		public override GridLines GridLines {
 			get {
-				if (!ControlStyleCreated) {
-#if NET_2_0
+				if (!ControlStyleCreated)
 					return GridLines.None;
-#else
-					return GridLines.Both;
-#endif
-				}
 				return TableStyle.GridLines;
 			}
 			set { TableStyle.GridLines = value; }
@@ -239,9 +213,8 @@ namespace System.Web.UI.WebControls {
 			get {
 				if (headerStyle == null) {
 					headerStyle = new TableItemStyle ();
-					if (IsTrackingViewState) {
+					if (IsTrackingViewState)
 						headerStyle.TrackViewState ();
-					}
 				}
 				return headerStyle;
 			}
@@ -280,9 +253,8 @@ namespace System.Web.UI.WebControls {
 			get {
 				if (itemStyle == null) {
 					itemStyle = new TableItemStyle ();
-					if (IsTrackingViewState) {
+					if (IsTrackingViewState)
 						itemStyle.TrackViewState ();
-					}
 				}
 				return itemStyle;
 			}
@@ -299,9 +271,6 @@ namespace System.Web.UI.WebControls {
 			set { itemTemplate = value; }
 		}
 
-#if ONLY_1_1
-		[Bindable (true)]
-#endif
 		[DefaultValue (0)]
 		[WebSysDescription ("")]
 		[WebCategory ("Layout")]
@@ -318,9 +287,6 @@ namespace System.Web.UI.WebControls {
 			}
 		}
 
-#if ONLY_1_1
-		[Bindable (true)]
-#endif
 		[DefaultValue (RepeatDirection.Vertical)]
 		[WebSysDescription ("")]
 		[WebCategory ("Layout")]
@@ -332,9 +298,6 @@ namespace System.Web.UI.WebControls {
 			set { ViewState ["RepeatDirection"] = value; }
 		}
 
-#if ONLY_1_1
-		[Bindable (true)]
-#endif
 		[DefaultValue (RepeatLayout.Table)]
 		[WebSysDescription ("")]
 		[WebCategory ("Layout")]
@@ -343,7 +306,13 @@ namespace System.Web.UI.WebControls {
 				object o = ViewState ["RepeatLayout"];
 				return (o == null) ? RepeatLayout.Table : (RepeatLayout) o;
 			}
-			set { ViewState ["RepeatLayout"] = value; }
+			set {
+#if NET_4_0
+				if (value == RepeatLayout.OrderedList || value == RepeatLayout.UnorderedList)
+					throw new ArgumentOutOfRangeException (String.Format ("DataList does not support the '{0}' layout.", value));
+#endif
+				ViewState ["RepeatLayout"] = value;
+			}
 		}
 
 		[Bindable (true)]
@@ -386,9 +355,8 @@ namespace System.Web.UI.WebControls {
 			get {
 				if (selectedItemStyle == null) {
 					selectedItemStyle = new TableItemStyle ();
-					if (IsTrackingViewState) {
+					if (IsTrackingViewState)
 						selectedItemStyle.TrackViewState ();
-					}
 				}
 				return selectedItemStyle;
 			}
@@ -415,9 +383,8 @@ namespace System.Web.UI.WebControls {
 			get {
 				if (separatorStyle == null) {
 					separatorStyle = new TableItemStyle ();
-					if (IsTrackingViewState) {
+					if (IsTrackingViewState)
 						separatorStyle.TrackViewState ();
-					}
 				}
 				return separatorStyle;
 			}
@@ -434,9 +401,6 @@ namespace System.Web.UI.WebControls {
 			set { separatorTemplate = value; }
 		}
 
-#if ONLY_1_1
-		[Bindable (true)]
-#endif
 		[DefaultValue (true)]
 		[WebSysDescription ("")]
 		[WebCategory ("Appearance")]
@@ -448,9 +412,6 @@ namespace System.Web.UI.WebControls {
 			set { ViewState ["ShowFooter"] = value; }
 		}
 
-#if ONLY_1_1
-		[Bindable (true)]
-#endif
 		[DefaultValue (true)]
 		[WebSysDescription ("")]
 		[WebCategory ("Appearance")]
@@ -462,7 +423,6 @@ namespace System.Web.UI.WebControls {
 			set { ViewState ["ShowHeader"] = value; }
 		}
 
-#if NET_2_0
 		[MonoTODO ("incomplete")]
 		[Browsable (false)]
 		public object SelectedValue {
@@ -470,9 +430,9 @@ namespace System.Web.UI.WebControls {
 				if (DataKeyField.Length == 0)
 					throw new InvalidOperationException (Locale.GetText ("No DataKeyField present."));
 
-				if ((SelectedIndex >= 0) && (selectedIndex < DataKeys.Count)) {
-					return DataKeys [selectedIndex];
-				}
+				int idx = SelectedIndex;
+				if ((idx >= 0) && (idx < DataKeys.Count))
+					return DataKeys [idx];
 
 				return null;
 			}
@@ -481,14 +441,13 @@ namespace System.Web.UI.WebControls {
 		protected override HtmlTextWriterTag TagKey {
 			get { return HtmlTextWriterTag.Table; }
 		}
-#endif
 
-		private TableStyle TableStyle {
+		TableStyle TableStyle {
 			// this will throw an InvalidCasException just like we need
 			get { return (TableStyle) ControlStyle; }
 		}
 		
-		private ArrayList ItemList {
+		ArrayList ItemList {
 			get {
 				if (list == null)
 					list = new ArrayList ();
@@ -521,7 +480,7 @@ namespace System.Web.UI.WebControls {
 			}
 		}
 
-		private void DoItemInLoop (int i, object d, bool databind, ListItemType type)
+		void DoItemInLoop (int i, object d, bool databind, ListItemType type)
 		{
 			DoItem (i, type, d, databind);
 			if (SeparatorTemplate != null)
@@ -532,23 +491,21 @@ namespace System.Web.UI.WebControls {
 		protected override void CreateControlHierarchy (bool useDataSource)
 		{
 			Controls.Clear();
+			ItemList.Clear ();
 
 			IEnumerable ds = null;
 			ArrayList keys = null;
 
 			if (useDataSource) {
 				idx = 0;
-#if NET_2_0
 				if (IsBoundUsingDataSourceID)
 					ds = GetData();
 				else
-#endif
-				ds = DataSourceResolver.ResolveDataSource (DataSource, DataMember);
+					ds = DataSourceResolver.ResolveDataSource (DataSource, DataMember);
 				keys = DataKeysArray;
 				keys.Clear ();
-			} else {
+			} else
 				idx = (int) ViewState ["Items"];
-			}
 
 			if ((ds == null) && (idx == 0))
 				return;
@@ -563,14 +520,14 @@ namespace System.Web.UI.WebControls {
 			if (ds != null) {
 				string key = DataKeyField;
 				foreach (object o in ds) {
-					if (useDataSource && key != "")
+					if (useDataSource && !String.IsNullOrEmpty (key))
 						keys.Add (DataBinder.GetPropertyValue (o, key));
 					type = ListItemType.Item;
 					if (idx == edit_item_index) 
 						type = ListItemType.EditItem;
 					else if (idx == selected_index) 
 						type = ListItemType.SelectedItem;
-					else if (idx % 2 != 0) 
+					else if ((idx & 1) != 0) 
 						type = ListItemType.AlternatingItem;
 
 					DoItemInLoop (idx, o, useDataSource, type);
@@ -579,11 +536,11 @@ namespace System.Web.UI.WebControls {
 			} else {
 				for (int i = 0; i < idx; i++) {
 					type = ListItemType.Item;
-					if (idx == edit_item_index) 
+					if (i == edit_item_index) 
 						type = ListItemType.EditItem;
-					else if (idx == selected_index) 
+					else if (i == selected_index) 
 						type = ListItemType.SelectedItem;
-					else if (idx % 2 != 0) 
+					else if ((i & 1) != 0) 
 						type = ListItemType.AlternatingItem;
 
 					DoItemInLoop (i, null, useDataSource, type);
@@ -598,12 +555,8 @@ namespace System.Web.UI.WebControls {
 
 		protected override Style CreateControlStyle ()
 		{
-#if NET_2_0
 			// not kept (directly) in the DataList ViewState
 			TableStyle tableStyle = new TableStyle ();
-#else
-			TableStyle tableStyle = new TableStyle (ViewState);
-#endif
 			tableStyle.CellSpacing = 0;
 			return tableStyle;
 		}
@@ -618,29 +571,28 @@ namespace System.Web.UI.WebControls {
 			ITemplate t = null;
 			
 			switch (item.ItemType) {
-			case ListItemType.Header:
-				t = HeaderTemplate;
-				break;
-			case ListItemType.Footer:
-				t = FooterTemplate;
-				break;	
-			case ListItemType.Separator:
-				t = SeparatorTemplate;
-				break;
-			case ListItemType.Item:
-			case ListItemType.AlternatingItem:
-			case ListItemType.SelectedItem:
-			case ListItemType.EditItem:
-				int index = item.ItemIndex;
-				if ((EditItemIndex == index) && (EditItemTemplate != null))
-					t = EditItemTemplate;
-				else if ((SelectedIndex == index) && (SelectedItemTemplate != null))
-					t = SelectedItemTemplate;
-				else if (((index & 1) == 0) || (alternatingItemTemplate == null))
-					t = ItemTemplate;
-				else
-					t = alternatingItemTemplate;
-				break;
+				case ListItemType.Header:
+					t = HeaderTemplate;
+					break;
+				case ListItemType.Footer:
+					t = FooterTemplate;
+					break;	
+				case ListItemType.Separator:
+					t = SeparatorTemplate;
+					break;
+				case ListItemType.Item:
+				case ListItemType.AlternatingItem:
+				case ListItemType.SelectedItem:
+				case ListItemType.EditItem:
+					if ((item.ItemType == ListItemType.EditItem) && (EditItemTemplate != null))
+						t = EditItemTemplate;
+					else if ((item.ItemType == ListItemType.SelectedItem) && (SelectedItemTemplate != null))
+						t = SelectedItemTemplate;
+					else if ((item.ItemType == ListItemType.AlternatingItem) && (AlternatingItemTemplate != null))
+						t = AlternatingItemTemplate;
+					else
+						t = ItemTemplate;
+					break;
 			}
 
 			if (t != null)
@@ -665,10 +617,8 @@ namespace System.Web.UI.WebControls {
 				HeaderStyle.LoadViewState (state [6]);
 			if (state [7] != null)
 				FooterStyle.LoadViewState (state [7]);
-#if NET_2_0
 			if (state [8] != null)
 				ControlStyle.LoadViewState (state [8]);
-#endif
 		}
 
 		protected override bool OnBubbleEvent (object source, EventArgs e)
@@ -678,21 +628,21 @@ namespace System.Web.UI.WebControls {
 				return false;
 
 			string cn = dlca.CommandName;
-			CultureInfo inv = CultureInfo.InvariantCulture;
+			CultureInfo inv = Helpers.InvariantCulture;
 
 			OnItemCommand (dlca);
-			if (String.Compare (cn, CancelCommandName, true, inv) == 0) {
+			if (String.Compare (cn, CancelCommandName, true, inv) == 0)
 				OnCancelCommand (dlca);
-			} else if (String.Compare (cn, DeleteCommandName, true, inv) == 0) {
+			else if (String.Compare (cn, DeleteCommandName, true, inv) == 0)
 				OnDeleteCommand (dlca);
-			} else if (String.Compare (cn, EditCommandName, true, inv) == 0) {
+			else if (String.Compare (cn, EditCommandName, true, inv) == 0)
 				OnEditCommand (dlca);
-			} else if (String.Compare (cn, SelectCommandName, true, inv) == 0) {
+			else if (String.Compare (cn, SelectCommandName, true, inv) == 0) {
 				SelectedIndex = dlca.Item.ItemIndex;
 				OnSelectedIndexChanged (dlca);
-			} else if (String.Compare (cn, UpdateCommandName, true, inv) == 0) {
+			} else if (String.Compare (cn, UpdateCommandName, true, inv) == 0)
 				OnUpdateCommand (dlca);
-			}
+			
 			return true;
 		}
 
@@ -716,14 +666,17 @@ namespace System.Web.UI.WebControls {
 			if (editCommand != null)
 				editCommand (this, e);
 		}
-#if NET_2_0
+
 		protected internal override void OnInit (EventArgs e)
 		{
-			// EditItemIndex and SelectedIndex now use the Control State (i.e not the View State)
-			Page.RegisterRequiresControlState (this);
+			// EditItemIndex and SelectedIndex now use the Control State (i.e not the
+			// View State)
+			Page page = Page;
+			if (page != null)
+				page.RegisterRequiresControlState (this);
 			base.OnInit (e);
 		}
-#endif
+
 		protected virtual void OnItemCommand (DataListCommandEventArgs e)
 		{
 			DataListCommandEventHandler itemCommand = (DataListCommandEventHandler) Events [itemCommandEvent];
@@ -760,72 +713,60 @@ namespace System.Web.UI.WebControls {
 			Style alt = null;
 			foreach (DataListItem item in Controls) {
 				switch (item.ItemType) {
-				case ListItemType.Item:
-					item.MergeStyle (itemStyle);
-					break;
-				case ListItemType.AlternatingItem:
-					if (alt == null) {
-						if (alternatingItemStyle != null) {
-							alt = new TableItemStyle ();
-							alt.CopyFrom (itemStyle);
-							alt.CopyFrom (alternatingItemStyle);
-						} else {
-							alt = itemStyle;
+					case ListItemType.Item:
+						item.MergeStyle (itemStyle);
+						break;
+					case ListItemType.AlternatingItem:
+						if (alt == null) {
+							if (alternatingItemStyle != null) {
+								alt = new TableItemStyle ();
+								alt.CopyFrom (itemStyle);
+								alt.CopyFrom (alternatingItemStyle);
+							} else
+								alt = itemStyle;
 						}
-					}
 
-					item.MergeStyle (alt);
-					break;
-				case ListItemType.EditItem:
-					if (editItemStyle != null) {
-						item.MergeStyle (editItemStyle);
-					} else {
-						item.MergeStyle (itemStyle);
-					}
-					break;
-				case ListItemType.Footer:
-					if (!ShowFooter) {
-						item.Visible = false;
+						item.MergeStyle (alt);
 						break;
-					}
-					if (footerStyle != null) {
-						item.MergeStyle (footerStyle);
-					}
-					break;
-				case ListItemType.Header:
-					if (!ShowHeader) {
-						item.Visible = false;
+					case ListItemType.EditItem:
+						if (editItemStyle != null)
+							item.MergeStyle (editItemStyle);
+						else
+							item.MergeStyle (itemStyle);
+						break;
+					case ListItemType.Footer:
+						if (!ShowFooter) {
+							item.Visible = false;
+							break;
+						}
+						if (footerStyle != null)
+							item.MergeStyle (footerStyle);
+						break;
+					case ListItemType.Header:
+						if (!ShowHeader) {
+							item.Visible = false;
+							break;
+						}
+						if (headerStyle != null)
+							item.MergeStyle (headerStyle);
+						break;
+					case ListItemType.SelectedItem:
+						if (selectedItemStyle != null)
+							item.MergeStyle (selectedItemStyle);
+						else
+							item.MergeStyle (itemStyle);
+						break;
+					case ListItemType.Separator:
+						if (separatorStyle != null)
+							item.MergeStyle(separatorStyle);
+						else
+							item.MergeStyle (itemStyle);
 						break;
-					}
-					if (headerStyle != null) {
-						item.MergeStyle (headerStyle);
-					}
-					break;
-				case ListItemType.SelectedItem:
-					if (selectedItemStyle != null) {
-						item.MergeStyle (selectedItemStyle);
-					} else {
-						item.MergeStyle (itemStyle);
-					}
-					break;
-				case ListItemType.Separator:
-					if (separatorStyle != null) {
-						item.MergeStyle(separatorStyle);
-					}
-					else {
-						item.MergeStyle (itemStyle);
-					}
-					break;
 				}
 			}
 		}
 
-#if NET_2_0
-		protected internal
-#else		
-		protected
-#endif		
-		override void RenderContents (HtmlTextWriter writer)
+		protected internal override void RenderContents (HtmlTextWriter writer)
 		{
 			if (Items.Count == 0)
 				return;			
@@ -856,24 +797,18 @@ for (int i=0; i < Items.Count; i++) {
 			if (extract) {
 				ri.OuterTableImplied = true;
 				writer.AddAttribute (HtmlTextWriterAttribute.Id, ClientID);
-				if (ControlStyleCreated) {
+				if (ControlStyleCreated)
 					ControlStyle.AddAttributesToRender (writer);
-				}
 				writer.RenderBeginTag (HtmlTextWriterTag.Table);
 				ri.RenderRepeater (writer, this, ControlStyle, this);
 				writer.RenderEndTag ();
-			} else {
+			} else
 				ri.RenderRepeater (writer, this, ControlStyle, this);
-			}
 		}
 
 		protected override object SaveViewState ()
 		{
-#if NET_2_0
 			object[] state = new object [9];
-#else
-			object[] state = new object [8];
-#endif
 			state[0] = base.SaveViewState ();
 			if (itemStyle != null)
 				state [1] = itemStyle.SaveViewState ();
@@ -889,10 +824,8 @@ for (int i=0; i < Items.Count; i++) {
 				state [6] = headerStyle.SaveViewState ();
 			if (footerStyle != null)
 				state [7] = footerStyle.SaveViewState ();
-#if NET_2_0
 			if (ControlStyleCreated)
 				state [8] = ControlStyle.SaveViewState ();
-#endif
 			return state;
 		}
 
@@ -913,10 +846,8 @@ for (int i=0; i < Items.Count; i++) {
 				selectedItemStyle.TrackViewState ();
 			if (separatorStyle != null)
 				separatorStyle.TrackViewState ();
-#if NET_2_0
 			if (ControlStyleCreated)
 				ControlStyle.TrackViewState ();
-#endif
 		}
 
 
@@ -999,31 +930,31 @@ for (int i=0; i < Items.Count; i++) {
 		{
 			DataListItem item = null;
 			switch (itemType) {
-			case ListItemType.Header:
-			case ListItemType.Footer:
-				if (repeatIndex >= 0 && (!HasControls () || repeatIndex >= Controls.Count))
-					throw new ArgumentOutOfRangeException ();
+				case ListItemType.Header:
+				case ListItemType.Footer:
+					if (repeatIndex >= 0 && (!HasControls () || repeatIndex >= Controls.Count))
+						throw new ArgumentOutOfRangeException ();
 
-				item = FindFirstItem (itemType);
-				break;
-			case ListItemType.Item:
-			case ListItemType.AlternatingItem:
-			case ListItemType.SelectedItem:
-			case ListItemType.EditItem:
-				if (repeatIndex >= 0 && (!HasControls () || repeatIndex >= Controls.Count))
-					throw new ArgumentOutOfRangeException ();
+					item = FindFirstItem (itemType);
+					break;
+				case ListItemType.Item:
+				case ListItemType.AlternatingItem:
+				case ListItemType.SelectedItem:
+				case ListItemType.EditItem:
+					if (repeatIndex >= 0 && (!HasControls () || repeatIndex >= Controls.Count))
+						throw new ArgumentOutOfRangeException ();
 
-				item = FindBestItem (repeatIndex);
-				break;
-			case ListItemType.Separator:
-				if (repeatIndex >= 0 && (!HasControls () || repeatIndex >= Controls.Count))
-					throw new ArgumentOutOfRangeException ();
+					item = FindBestItem (repeatIndex);
+					break;
+				case ListItemType.Separator:
+					if (repeatIndex >= 0 && (!HasControls () || repeatIndex >= Controls.Count))
+						throw new ArgumentOutOfRangeException ();
 
-				item = FindSpecificItem (itemType, repeatIndex);
-				break;
-			default:
-				item = null;
-				break;
+					item = FindSpecificItem (itemType, repeatIndex);
+					break;
+				default:
+					item = null;
+					break;
 			}
 
 			if (item == null || item.ControlStyleCreated == false)
@@ -1033,7 +964,7 @@ for (int i=0; i < Items.Count; i++) {
 		}
 
 		// Header and Footer don't have a "real" index (-1)
-		private DataListItem FindFirstItem (ListItemType itemType)
+		DataListItem FindFirstItem (ListItemType itemType)
 		{
 			for (int i = 0; i < Controls.Count; i++) {
 				DataListItem item = (Controls [i] as DataListItem);
@@ -1044,7 +975,7 @@ for (int i=0; i < Items.Count; i++) {
 		}
 
 		// Both Type and Index must match (e.g. Separator)
-		private DataListItem FindSpecificItem (ListItemType itemType, int repeatIndex)
+		DataListItem FindSpecificItem (ListItemType itemType, int repeatIndex)
 		{
 			for (int i = 0; i < Controls.Count; i++) {
 				DataListItem item = (Controls [i] as DataListItem);
@@ -1055,19 +986,19 @@ for (int i=0; i < Items.Count; i++) {
 		}
 
 		// we get call for Item even for AlternatingItem :(
-		private DataListItem FindBestItem (int repeatIndex)
+		DataListItem FindBestItem (int repeatIndex)
 		{
 			for (int i = 0; i < Controls.Count; i++) {
 				DataListItem item = (Controls [i] as DataListItem);
 				if ((item != null) && (item.ItemIndex == repeatIndex)) {
 					switch (item.ItemType) {
-					case ListItemType.Item:
-					case ListItemType.AlternatingItem:
-					case ListItemType.SelectedItem:
-					case ListItemType.EditItem:
-						return item;
-					default:
-						return null;
+						case ListItemType.Item:
+						case ListItemType.AlternatingItem:
+						case ListItemType.SelectedItem:
+						case ListItemType.EditItem:
+							return item;
+						default:
+							return null;
 					}
 				}
 			}
@@ -1082,19 +1013,19 @@ for (int i=0; i < Items.Count; i++) {
 
 			DataListItem item = null;
 			switch (itemType) {
-			case ListItemType.Header:
-			case ListItemType.Footer:
-				item = FindFirstItem (itemType);
-				break;
-			case ListItemType.Item:
-			case ListItemType.AlternatingItem:
-			case ListItemType.SelectedItem:
-			case ListItemType.EditItem:
-				item = FindBestItem (repeatIndex);
-				break;
-			case ListItemType.Separator:
-				item = FindSpecificItem (itemType, repeatIndex);
-				break;
+				case ListItemType.Header:
+				case ListItemType.Footer:
+					item = FindFirstItem (itemType);
+					break;
+				case ListItemType.Item:
+				case ListItemType.AlternatingItem:
+				case ListItemType.SelectedItem:
+				case ListItemType.EditItem:
+					item = FindBestItem (repeatIndex);
+					break;
+				case ListItemType.Separator:
+					item = FindSpecificItem (itemType, repeatIndex);
+					break;
 			}
 
 			if (item != null) {
