@@ -198,7 +198,7 @@ namespace System.Configuration
 		}
 		internal static string GetMachineConfigPath ()
 		{
-			return "/META-INF/machine.config";
+			return System.Runtime.InteropServices.RuntimeEnvironment.SystemConfigurationFile;
 		}
 #else
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
@@ -240,7 +240,9 @@ namespace System.Configuration
 		public readonly string TypeName;
 		public readonly bool AllowLocation;
 		public readonly AllowDefinition AllowDefinition;
+#if XML_DEP
 		public string FileName;
+#endif
 		public readonly bool RequirePermission;
 
 		public SectionData (string sectionName, string typeName,
@@ -742,7 +744,7 @@ namespace System.Configuration
 
 		private void ReadConfigFile (XmlTextReader reader)
 		{
-			int depth = reader.Depth;
+			//int depth = reader.Depth;
 			for (reader.MoveToContent ();
 			     !reader.EOF && reader.NodeType != XmlNodeType.EndElement;
 			     reader.MoveToContent ()) {
