@@ -22,7 +22,7 @@
 //
 //
 
-#if NET_4_0 || BOOTSTRAP_NET_4_0
+#if NET_4_0 || MOBILE
 
 using System;
 using System.Threading;
@@ -51,8 +51,7 @@ namespace System.Threading.Tasks
 		int                 taskId;
 		TaskCreationOptions taskCreationOptions;
 		
-		IScheduler          scheduler;
-		TaskScheduler       taskScheduler;
+		TaskScheduler       scheduler;
 
 		ManualResetEventSlim schedWait = new ManualResetEventSlim (false);
 		
@@ -63,6 +62,7 @@ namespace System.Threading.Tasks
 		
 		Action<object> action;
 		object         state;
+		AtomicBooleanValue executing;
 
 		ConcurrentQueue<EventHandler> completed = new ConcurrentQueue<EventHandler> ();
 
@@ -133,7 +133,7 @@ namespace System.Threading.Tasks
 		static void EmptyFunc (object o)
 		{
 		}
-		
+
 		#region Start
 		public void Start ()
 		{
@@ -141,34 +141,18 @@ namespace System.Threading.Tasks
 		}
 		
 		public void Start (TaskScheduler scheduler)
-		{
-			Start (ProxifyScheduler (scheduler));
-		}
-		
-		void Start (IScheduler scheduler)
 		{
 			SetupScheduler (scheduler);
 			Schedule ();
 		}
 
-		internal void SetupScheduler (TaskScheduler tscheduler)
-		{
-			SetupScheduler (ProxifyScheduler (tscheduler));
-		}
-
-		internal void SetupScheduler (IScheduler scheduler)
+		internal void SetupScheduler (TaskScheduler scheduler)
 		{
 			this.scheduler = scheduler;
 			status = TaskStatus.WaitingForActivation;
 			schedWait.Set ();
 		}
 		
-		IScheduler ProxifyScheduler (TaskScheduler tscheduler)
-		{
-			IScheduler sched = tscheduler as IScheduler;
-			return sched != null ? sched : new SchedulerProxy (tscheduler);
-		}
-		
 		public void RunSynchronously ()
 		{
 			RunSynchronously (TaskScheduler.Current);
@@ -209,6 +193,7 @@ namespace System.Threading.Tasks
 		{
 			Task continuation = new Task ((o) => continuationAction ((Task)o), this, cancellationToken, GetCreationOptions (continuationOptions));
 			ContinueWithCore (continuation, continuationOptions, scheduler);
+
 			return continuation;
 		}
 		
@@ -256,8 +241,7 @@ namespace System.Threading.Tasks
 		                                TaskScheduler scheduler, Func<bool> predicate)
 		{
 			// Already set the scheduler so that user can call Wait and that sort of stuff
-			continuation.taskScheduler = scheduler;
-			continuation.scheduler = ProxifyScheduler (scheduler);
+			continuation.scheduler = scheduler;
 			continuation.schedWait.Set ();
 			continuation.status = TaskStatus.WaitingForActivation;
 			
@@ -299,6 +283,12 @@ namespace System.Threading.Tasks
 			int kindCode = (int)kind;
 			
 			if (kindCode >= ((int)TaskContinuationOptions.NotOnRanToCompletion)) {
+				// Remove other options
+				kind &= ~(TaskContinuationOptions.PreferFairness
+				          | TaskContinuationOptions.LongRunning
+				          | TaskContinuationOptions.AttachedToParent
+				          | TaskContinuationOptions.ExecuteSynchronously);
+
 				if (status == TaskStatus.Canceled) {
 					if (kind == TaskContinuationOptions.NotOnCanceled)
 						return false;
@@ -328,8 +318,7 @@ namespace System.Threading.Tasks
 		
 		void CheckAndSchedule (Task continuation, TaskContinuationOptions options, TaskScheduler scheduler, bool fromCaller)
 		{
-			if (!fromCaller 
-			    && (options & TaskContinuationOptions.ExecuteSynchronously) > 0)
+			if ((options & TaskContinuationOptions.ExecuteSynchronously) > 0)
 				continuation.ThreadStart ();
 			else
 				continuation.Start (scheduler);
@@ -355,8 +344,9 @@ namespace System.Threading.Tasks
 			status = TaskStatus.WaitingToRun;
 			
 			// If worker is null it means it is a local one, revert to the old behavior
-			if (childWorkAdder == null || CheckTaskOptions (taskCreationOptions, TaskCreationOptions.PreferFairness)) {
-				scheduler.AddWork (this);
+			// If TaskScheduler.Current is not being used, the scheduler was explicitly provided, so we must use that
+			if (scheduler != TaskScheduler.Current || childWorkAdder == null || CheckTaskOptions (taskCreationOptions, TaskCreationOptions.PreferFairness)) {
+				scheduler.QueueTask (this);
 			} else {
 				/* Like the semantic of the ABP paper describe it, we add ourselves to the bottom 
 				 * of our Parent Task's ThreadWorker deque. It's ok to do that since we are in
@@ -368,8 +358,15 @@ namespace System.Threading.Tasks
 		
 		void ThreadStart ()
 		{
+			/* Allow scheduler to break fairness of deque ordering without
+			 * breaking its semantic (the task can be executed twice but the
+			 * second time it will return immediately
+			 */
+			if (!executing.TryRelaxedSet ())
+				return;
+
 			current = this;
-			TaskScheduler.Current = taskScheduler;
+			TaskScheduler.Current = scheduler;
 			
 			if (!token.IsCancellationRequested) {
 				
@@ -447,8 +444,6 @@ namespace System.Threading.Tasks
 			if (CheckTaskOptions (taskCreationOptions, TaskCreationOptions.AttachedToParent) && parent != null){
 				parent.ChildCompleted ();
 			}
-			
-			Dispose ();
 		}
 
 		void ProcessCompleteDelegates ()
@@ -475,7 +470,7 @@ namespace System.Threading.Tasks
 		{
 			exception = e;
 			status = TaskStatus.Faulted;
-			if (taskScheduler != null && taskScheduler.FireUnobservedEvent (exception).Observed)
+			if (scheduler != null && scheduler.FireUnobservedEvent (exception).Observed)
 				exceptionObserved = true;
 		}
 		
@@ -524,9 +519,13 @@ namespace System.Threading.Tasks
 				millisecondsTimeout = ComputeTimeout (millisecondsTimeout, watch);
 			}
 
-			Func<bool> stopFunc
-				= delegate { cancellationToken.ThrowIfCancellationRequested (); return watch.ElapsedMilliseconds > millisecondsTimeout; };
-			bool result = scheduler.ParticipateUntil (this, stopFunc);
+			ManualResetEventSlim predicateEvt = new ManualResetEventSlim (false);
+			if (cancellationToken != CancellationToken.None) {
+				cancellationToken.Register (predicateEvt.Set);
+				cancellationToken.ThrowIfCancellationRequested ();
+			}
+
+			bool result = scheduler.ParticipateUntil (this, predicateEvt, millisecondsTimeout);
 
 			if (exception != null)
 				throw exception;
@@ -540,34 +539,39 @@ namespace System.Threading.Tasks
 		{
 			if (tasks == null)
 				throw new ArgumentNullException ("tasks");
-			if (tasks.Length == 0)
-				throw new ArgumentException ("tasks is empty", "tasks");
 			
-			foreach (var t in tasks)
+			foreach (var t in tasks) {
+				if (t == null)
+					throw new ArgumentNullException ("tasks", "the tasks argument contains a null element");
 				t.Wait ();
+			}
 		}
 
 		public static void WaitAll (Task[] tasks, CancellationToken cancellationToken)
 		{
 			if (tasks == null)
 				throw new ArgumentNullException ("tasks");
-			if (tasks.Length == 0)
-				throw new ArgumentException ("tasks is empty", "tasks");
 			
-			foreach (var t in tasks)
+			foreach (var t in tasks) {
+				if (t == null)
+					throw new ArgumentNullException ("tasks", "the tasks argument contains a null element");
+
 				t.Wait (cancellationToken);
+			}
 		}
 		
 		public static bool WaitAll (Task[] tasks, TimeSpan timeout)
 		{
 			if (tasks == null)
 				throw new ArgumentNullException ("tasks");
-			if (tasks.Length == 0)
-				throw new ArgumentException ("tasks is empty", "tasks");
 			
 			bool result = true;
-			foreach (var t in tasks)
+			foreach (var t in tasks) {
+				if (t == null)
+					throw new ArgumentNullException ("tasks", "the tasks argument contains a null element");
+
 				result &= t.Wait (timeout);
+			}
 			return result;
 		}
 		
@@ -575,12 +579,14 @@ namespace System.Threading.Tasks
 		{
 			if (tasks == null)
 				throw new ArgumentNullException ("tasks");
-			if (tasks.Length == 0)
-				throw new ArgumentException ("tasks is empty", "tasks");
 			
 			bool result = true;
-			foreach (var t in tasks)
+			foreach (var t in tasks) {
+				if (t == null)
+					throw new ArgumentNullException ("tasks", "the tasks argument contains a null element");
+
 				result &= t.Wait (millisecondsTimeout);
+			}
 			return result;
 		}
 		
@@ -588,12 +594,14 @@ namespace System.Threading.Tasks
 		{
 			if (tasks == null)
 				throw new ArgumentNullException ("tasks");
-			if (tasks.Length == 0)
-				throw new ArgumentException ("tasks is empty", "tasks");
 			
 			bool result = true;
-			foreach (var t in tasks)
+			foreach (var t in tasks) {
+				if (t == null)
+					throw new ArgumentNullException ("tasks", "the tasks argument contains a null element");
+
 				result &= t.Wait (millisecondsTimeout, cancellationToken);
+			}
 			return result;
 		}
 		
@@ -637,8 +645,10 @@ namespace System.Threading.Tasks
 			int numFinished = 0;
 			int indexFirstFinished = -1;
 			int index = 0;
-			IScheduler sched = null;
+			TaskScheduler sched = null;
+			Task task = null;
 			Watch watch = Watch.StartNew ();
+			ManualResetEventSlim predicateEvt = new ManualResetEventSlim (false);
 
 			foreach (Task t in tasks) {
 				int indexResult = index++;
@@ -650,10 +660,15 @@ namespace System.Threading.Tasks
 					// Check if we are the first to have finished
 					if (result == 1)
 						indexFirstFinished = indexResult;
+
+					// Stop waiting
+					predicateEvt.Set ();
 				}, TaskContinuationOptions.ExecuteSynchronously);
 
-				if (sched == null && t.scheduler != null)
+				if (sched == null && t.scheduler != null) {
+					task = t;
 					sched = t.scheduler;
+				}
 			}
 
 			// If none of task have a scheduler we are forced to wait for at least one to start
@@ -663,22 +678,20 @@ namespace System.Threading.Tasks
 				if ((shandle = WaitHandle.WaitAny (handles, millisecondsTimeout)) == WaitHandle.WaitTimeout)
 					return -1;
 				sched = tasks[shandle].scheduler;
+				task = tasks[shandle];
 				millisecondsTimeout = ComputeTimeout (millisecondsTimeout, watch);
 			}
-			
+
 			// One task already finished
 			if (indexFirstFinished != -1)
 				return indexFirstFinished;
-			
-			// All tasks are supposed to use the same TaskScheduler
-			sched.ParticipateUntil (delegate {
-				if (millisecondsTimeout != -1 && watch.ElapsedMilliseconds > millisecondsTimeout)
-					return true;
 
+			if (cancellationToken != CancellationToken.None) {
+				cancellationToken.Register (predicateEvt.Set);
 				cancellationToken.ThrowIfCancellationRequested ();
+			}
 
-				return numFinished >= 1;
-			});
+			sched.ParticipateUntil (task, predicateEvt, millisecondsTimeout);
 
 			// Index update is still not done
 			if (indexFirstFinished == -1) {
@@ -686,7 +699,7 @@ namespace System.Threading.Tasks
 				while (indexFirstFinished == -1)
 					wait.SpinOnce ();
 			}
-			
+
 			return indexFirstFinished;
 		}
 
@@ -718,7 +731,6 @@ namespace System.Threading.Tasks
 			// any big object references that the user might have captured in a anonymous method
 			if (disposing) {
 				action = null;
-				completed.Clear ();
 				state = null;
 			}
 		}
@@ -806,6 +818,12 @@ namespace System.Threading.Tasks
 				return taskId;
 			}
 		}
+
+		internal Task Parent {
+			get {
+				return parent;
+			}
+		}
 		#endregion
 	}
 }
