@@ -71,7 +71,7 @@
 
 #include "mono/io-layer/socket-wrappers.h"
 
-#ifdef HOST_WIN32
+#if defined(HOST_WIN32)
 /* This is a kludge to make this file build under cygwin:
  * w32api/ws2tcpip.h has definitions for some AF_INET6 values and
  * prototypes for some but not all required functions (notably
@@ -553,6 +553,7 @@ static gint32 convert_sockopt_level_and_name(MonoSocketOptionLevel mono_level,
 
 		switch(mono_name) {
 		case SocketOptionName_IpTimeToLive:
+		case SocketOptionName_HopLimit:
 			*system_name = IPV6_UNICAST_HOPS;
 			break;
 		case SocketOptionName_MulticastInterface:
@@ -2805,6 +2806,7 @@ addrinfo_to_IPHostEntry(struct addrinfo *info, MonoString **h_name,
 		mono_array_setref (*h_addr_list, addr_index, addr_string);
 
 		if(!i) {
+			i++;
 			if (ai->ai_canonname != NULL) {
 				*h_name=mono_string_new(domain, ai->ai_canonname);
 			} else {
@@ -2837,19 +2839,23 @@ MonoBoolean ves_icall_System_Net_Dns_GetHostByName_internal(MonoString *host, Mo
 	MONO_ARCH_SAVE_REGS;
 	
 	hostname=mono_string_to_utf8 (host);
-	if (*hostname == '\0')
+	if (*hostname == '\0') {
 		add_local_ips = TRUE;
+		*h_name = host;
+	}
 #ifdef HAVE_SIOCGIFCONF
 	if (!add_local_ips && gethostname (this_hostname, sizeof (this_hostname)) != -1) {
-		if (!strcmp (hostname, this_hostname))
+		if (!strcmp (hostname, this_hostname)) {
 			add_local_ips = TRUE;
+			*h_name = host;
+		}
 	}
 #endif
 
 	memset(&hints, 0, sizeof(hints));
 	hints.ai_family = get_family_hint ();
 	hints.ai_socktype = SOCK_STREAM;
-	hints.ai_flags = AI_CANONNAME | AI_ADDRCONFIG;
+	hints.ai_flags = AI_CANONNAME;
 
 	if (*hostname && getaddrinfo(hostname, NULL, &hints, &info) == -1) {
 		return(FALSE);
