@@ -1486,7 +1486,7 @@ class Tests {
 			val = d / q;
 		} catch (DivideByZeroException) {
 			/* wrong exception */
-		} catch (ArithmeticException) {
+		} catch (OverflowException) {
 			failed = false;
 		}
 		if (failed)
@@ -1499,7 +1499,7 @@ class Tests {
 			val = d % q;
 		} catch (DivideByZeroException) {
 			/* wrong exception */
-		} catch (ArithmeticException) {
+		} catch (OverflowException) {
 			failed = false;
 		}
 		if (failed)
@@ -2240,6 +2240,8 @@ class Tests {
 
 		public static void rethrow2 () {
 			rethrow1 ();
+			/* This disables tailcall opts */
+			Console.WriteLine ();
 		}
 	}
 
